import { TestingModule, Test } from '@nestjs/testing';
import { DatabaseModule } from '../../../database/database.module';
import { PrismaService } from '../../../database/adapters/secondaries/prisma-service';
import { DatabaseException } from '../../../database/exceptions/database.exception';
import { TerritoriesRepository } from '../../adapters/secondaries/territories.repository';
import { v4 as uuidv4 } from 'uuid';
import { Territory } from '../../domain/entities/territory';
import { Point } from '../../domain/entities/point';

describe('TerritoriesRepository', () => {
  let prismaService: PrismaService;
  let territoriesRepository: TerritoriesRepository;

  const createTerritories = async (nbToCreate = 10) => {
    for (let i = 0; i < nbToCreate; i++) {
      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('${uuidv4()}','name-${i}',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 0, 0 ], [ 0, ${
        (i + 1) * 2
      } ], [ ${(i + 1) * 2}, ${(i + 1) * 2} ], [ ${
        (i + 1) * 2
      }, 0 ], [ 0, 0 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);
    }
  };

  const createExtraTerritory = async () => {
    const command = `INSERT INTO territory (uuid, name, shape) VALUES ('${uuidv4()}','name-extra',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 100, 100 ], [ 100, 110 ], [ 110, 110 ], [ 110, 100 ], [ 100, 100 ] ] ] ]}'))`;
    await prismaService.$executeRawUnsafe(command);
  };

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [DatabaseModule],
      providers: [TerritoriesRepository, PrismaService],
    }).compile();

    prismaService = module.get<PrismaService>(PrismaService);
    territoriesRepository = module.get<TerritoriesRepository>(
      TerritoriesRepository,
    );
  });

  afterAll(async () => {
    await prismaService.$disconnect();
  });

  beforeEach(async () => {
    await prismaService.territory.deleteMany();
  });

  describe('findAll', () => {
    it('should return an empty data array', async () => {
      const res = await territoriesRepository.findAll();
      expect(res).toEqual({
        data: [],
        total: 0,
      });
    });

    it('should return a data array with 8 territories', async () => {
      await createTerritories(8);
      const users = await territoriesRepository.findAll();
      expect(users.data.length).toBe(8);
      expect(users.total).toBe(8);
    });

    it('should return a data array limited to 10 territories', async () => {
      await createTerritories(20);
      const users = await territoriesRepository.findAll();
      expect(users.data.length).toBe(10);
      expect(users.total).toBe(20);
    });
  });

  describe('findAllForPoint', () => {
    it('should return a data array with 3 territories', async () => {
      await createTerritories(10);
      const users = await territoriesRepository.findForPoint(new Point(15, 15));
      expect(users.data.length).toBe(3);
      expect(users.total).toBe(3);
    });

    it('should return an empty data array', async () => {
      await createTerritories(10);
      const users = await territoriesRepository.findForPoint(new Point(50, 50));
      expect(users.data.length).toBe(0);
      expect(users.total).toBe(0);
    });
  });

  describe('findAllForPoints', () => {
    it('should return a data array with 4 territories', async () => {
      await createTerritories(10);
      await createExtraTerritory();
      const users = await territoriesRepository.findForPoints([
        new Point(15, 15),
        new Point(105, 105),
      ]);
      expect(users.data.length).toBe(4);
      expect(users.total).toBe(4);
    });
  });

  describe('findOneByUuid', () => {
    it('should return a territory', async () => {
      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('d83a1697-fc82-4f84-b7f4-dd9e2bcae327','test',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 4.7953936, 46.2200671 ], [ 4.793836, 46.2155098 ], [ 4.7931368, 46.2124124 ], [ 4.7927677, 46.207594 ], [ 4.793866, 46.2042467 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);

      const territory = await territoriesRepository.findOneByUuid(
        'd83a1697-fc82-4f84-b7f4-dd9e2bcae327',
      );
      expect(territory.name).toBe('test');
    });

    it('should return null', async () => {
      const territory = await territoriesRepository.findOneByUuid(
        '544572be-11fb-4244-8235-587221fc9104',
      );
      expect(territory).toBeNull();
    });
  });

  describe('findOne', () => {
    it('should return a territory according to its name', async () => {
      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('d83a1697-fc82-4f84-b7f4-dd9e2bcae327','test',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 4.7953936, 46.2200671 ], [ 4.793836, 46.2155098 ], [ 4.7931368, 46.2124124 ], [ 4.7927677, 46.207594 ], [ 4.793866, 46.2042467 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);

      const territory = await territoriesRepository.findOne({
        name: 'test',
      });

      expect(territory.uuid).toBe('d83a1697-fc82-4f84-b7f4-dd9e2bcae327');
    });

    it('should return null with unknown name', async () => {
      const territory = await territoriesRepository.findOne({
        name: 'wrong name',
      });
      expect(territory).toBeNull();
    });
  });

  describe('create', () => {
    it('should create a territory', async () => {
      const beforeCount = await prismaService.territory.count();

      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('d83a1697-fc82-4f84-b7f4-dd9e2bcae327','test',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 4.7953936, 46.2200671 ], [ 4.793836, 46.2155098 ], [ 4.7931368, 46.2124124 ], [ 4.7927677, 46.207594 ], [ 4.793866, 46.2042467 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);

      const afterCount = await prismaService.territory.count();

      expect(afterCount - beforeCount).toBe(1);
    });
  });

  describe('update', () => {
    it('should update territory name', async () => {
      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('d83a1697-fc82-4f84-b7f4-dd9e2bcae327','test',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 4.7953936, 46.2200671 ], [ 4.793836, 46.2155098 ], [ 4.7931368, 46.2124124 ], [ 4.7927677, 46.207594 ], [ 4.793866, 46.2042467 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);

      const toUpdate: Territory = new Territory();
      toUpdate.name = "'updated'";
      await territoriesRepository.updateWithFields(
        'd83a1697-fc82-4f84-b7f4-dd9e2bcae327',
        toUpdate,
      );

      const updatedTerritory = await territoriesRepository.findOneByUuid(
        'd83a1697-fc82-4f84-b7f4-dd9e2bcae327',
      );
      expect(updatedTerritory.name).toBe('updated');
    });

    it('should return 0 updated territories', async () => {
      const toUpdate: Territory = new Territory();
      toUpdate.name = "'updated'";

      const affectedRowNumber = await territoriesRepository.updateWithFields(
        '544572be-11fb-4244-8235-587221fc9104',
        toUpdate,
      );
      expect(affectedRowNumber).toBe(0);
    });
  });

  describe('delete', () => {
    it('should delete a territory', async () => {
      const command = `INSERT INTO territory (uuid, name, shape) VALUES ('d83a1697-fc82-4f84-b7f4-dd9e2bcae327','test',ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":[ [ [ [ 4.7953936, 46.2200671 ], [ 4.793836, 46.2155098 ], [ 4.7931368, 46.2124124 ], [ 4.7927677, 46.207594 ], [ 4.793866, 46.2042467 ] ] ] ]}'))`;
      await prismaService.$executeRawUnsafe(command);

      await territoriesRepository.delete(
        'd83a1697-fc82-4f84-b7f4-dd9e2bcae327',
      );

      const count = await prismaService.territory.count();
      expect(count).toBe(0);
    });

    it('should throw DatabaseException', async () => {
      await expect(
        territoriesRepository.delete('544572be-11fb-4244-8235-587221fc9104'),
      ).rejects.toBeInstanceOf(DatabaseException);
    });
  });
});
